import socket
import time

import requests
from zeroconf import ServiceBrowser, ServiceInfo, ServiceStateChange, Zeroconf


def discover(timeout: int = 3) -> list[str]:
    # Discover WLED instances on the local network
    def check_wled_on_ip(info: ServiceInfo):
        address = socket.inet_ntoa(info.addresses[0])
        port = info.port
        try:
            response = requests.get(f"http://{address}:{port}/win", timeout=0.1)
            if response.status_code == 200:
                services.append(address)
        except requests.exceptions.RequestException as e:
            pass  # Service didn't respond to /win or there was another error

    zeroconf = Zeroconf()
    services = []

    def on_service_state_change(zeroconf, service_type, name, state_change):
        if state_change == ServiceStateChange.Added:
            info = zeroconf.get_service_info(service_type, name)
            if info:
                check_wled_on_ip(info)

    try:
        with ServiceBrowser(
            zeroconf, "_http._tcp.local.", handlers=[on_service_state_change]
        ) as browser:
            time.sleep(timeout)
    finally:  # Clean up
        zeroconf.close()

    if not services:
        print("No WLED instances found on the network.")
        exit(1)
    return services


def select_instance(instances: list[str]) -> str | None:
    # Ask user which WLED instance to cast to
    if not instances:
        print("No WLED instances available to select.")
        return None
    print("Please select a WLED instance to cast to:")
    for i, instance in enumerate(instances):
        print(f"{i+1}. {instance}")
    selected_index = int(input("Enter the number of your selection: ")) - 1
    if selected_index < 0 or selected_index >= len(instances):
        print("Invalid selection. Please try again.")
        return select_instance(instances)

    return instances[selected_index]


def get_matrix_shape(host) -> tuple[int, int]:
    # Determine the shape of the LED pixel matrix from WLED
    try:
        response = requests.get(f"http://{host}:80/json/state")
        data = response.json()
        seg_data = data.get("seg")
        if seg_data is None or not seg_data:
            print(
                f"Error: The WLED instance did not return the expected 'seg' data. Response data: {data}"
            )
            exit(1)
        # Find the first segment that is turned on
        on_segment = next((seg for seg in seg_data if seg.get("on")), None)
        if on_segment is None:
            print(
                f"Error: Make sure you have at least one segment turned on and try again. Response data: {data}"
            )
            exit(1)
        width = on_segment.get("stop") - on_segment.get("start")
        try:
            height = on_segment.get("stopY") - on_segment.get("startY")
        except KeyError:
            height = 1
        return (width, height)
    except requests.exceptions.RequestException as e:
        print(f"Error communicating with the WLED instance: {e}")
        exit(1)



import json
import os
import cv2
import numpy as np
import wx
from PIL import Image, ImageEnhance

with open(os.path.join(os.path.dirname(__file__), "filter.json")) as f:
    config = json.load(f)

def apply_filters_pillow(image: Image):
    filters = config
    brightness = ImageEnhance.Brightness(image)
    image = brightness.enhance(filters["brightness"])
    color = ImageEnhance.Color(image)
    image = color.enhance(filters["saturation"])
    contrast = ImageEnhance.Contrast(image)
    image = contrast.enhance(filters["contrast"])
    sharpness = ImageEnhance.Sharpness(image)
    image = sharpness.enhance(filters["sharpen"])
    return image


def apply_filters_cv2(img: np.ndarray) -> np.ndarray:
    filters = config
    # Convert to HSV for color adjustment
    if filters["saturation"] is not None:
        img = filter_saturation(img, filters["saturation"])

    # Adjust brightness
    if filters["brightness"] is not None:
        img = filter_brightness(img, filters["brightness"])

    # Adjust contrast
    if filters["contrast"] is not None:
        img = filter_contrast(img, filters["contrast"])

    if filters["sharpen"] is not None:
        img = filter_sharpen(img, filters["sharpen"])

    if filters["balance_r"] is not None:
        img = filter_balance(
            img,
            {
                "r": filters["balance_r"],
                "g": filters["balance_g"],
                "b": filters["balance_b"],
            },
        )

    return img


def filter_sharpen(img, alpha):
    kernel = np.array([[0, -1, 0], [-1, 4, -1], [0, -1, 0]]) * alpha
    kernel[1, 1] += 1
    img = cv2.filter2D(img, -1, kernel)
    return img


def filter_balance(img, alpha):
    # scale the red, green, and blue channels
    scale = np.array([alpha["r"], alpha["g"], alpha["b"]])[np.newaxis, np.newaxis, :]

    img = (img * scale).astype(np.uint8)
    return img


def filter_contrast(img, alpha):
    # Compute the mean luminance (gray level)
    mean_luminance = np.mean(img)

    # Create a gray image of mean luminance
    gray_img = np.full_like(img, mean_luminance)

    # Enhance contrast
    enhanced_img = cv2.addWeighted(img, alpha, gray_img, 1 - alpha, 0)
    return enhanced_img


def filter_brightness(img, alpha):
    # Create a black image
    black_img = np.zeros_like(img)

    # Enhance brightness
    enhanced_img = cv2.addWeighted(img, alpha, black_img, 1 - alpha, 0)
    return enhanced_img


def filter_saturation(img, alpha):
    # Convert to HSV and split the channels
    hsv = cv2.cvtColor(img, cv2.COLOR_RGB2HSV)
    h, s, v = cv2.split(hsv)

    # Create a grayscale (desaturated) version
    gray = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)

    # Enhance color
    s_enhanced = cv2.addWeighted(s, alpha, gray, 1 - alpha, 0)

    # Merge and convert back to RGB
    enhanced_img = cv2.cvtColor(cv2.merge([h, s_enhanced, v]), cv2.COLOR_HSV2RGB)
    return enhanced_img


def image_to_ascii(image):
    # Convert the image to ASCII art
    ascii_chars = "@%#*+=-:. "
    width, height = image.size
    image = image.resize((width, height // 2))  # Correct aspect ratio
    image = image.convert("L")  # Convert to grayscale
    pixels = image.getdata()
    ascii_str = ""
    for pixel_value in pixels:
        ascii_str += ascii_chars[
            pixel_value // 32
        ]  # Map the pixel value to ascii_chars
    ascii_str_len = len(ascii_str)
    ascii_img = ""
    for i in range(0, ascii_str_len, width):
        ascii_img += ascii_str[i : i + width] + "\n"
    return ascii_img

def stretch_array_repeat(arr, stretch_factor):
    # Repeat the array elements along the column axis
    stretched_arr = np.repeat(arr, stretch_factor, axis=0)
    # Repeat the array elements along the row axis
    stretched_arr = np.repeat(stretched_arr, stretch_factor, axis=1)
    return stretched_arr

def show_preview(rgb_array, resolution):
    # Show a live view of the downscaled feed on the computer with 24-bit color blocks
    # Scaling by an integer will keep the pixelated effect but make it visible
    x, y = resolution
    monitor_x, monitor_y = wx.GetDisplaySize()
    # find the largest integer scaling factpr that will fit in the monitor in both dimensions
    scaling_factor = min(monitor_x // x, monitor_y // y)
    bgr_array = cv2.cvtColor(rgb_array, cv2.COLOR_RGB2BGR)
    if scaling_factor > 1:
        bgr_array = stretch_array_repeat(bgr_array, scaling_factor)
    try:
        # Use OpenCV to display the image
        cv2.imshow("Live View", bgr_array)
        cv2.waitKey(1)

    except Exception as e:
        print(f"An error occurred during live view: {e}")



import socket

import numpy


class PixelWriter:
    DDP_MAX_DATALEN = 1200  # Maximum length of DDP data
    DDP_DESTINATION_ID = 1  # Hardcoded Destination ID

    def __init__(self, host):
        self.host = host
        self.sequence_id = 1  # Initialize sequence ID
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  # UDP socket

    def update_pixels(self, rgb_array: numpy.ndarray):
        # Update the LED matrix via WLED in real-time using DDP
        flat_screen_data = rgb_array.flatten().tolist()
        # Convert the list to a bytearray
        byte_data = bytearray(flat_screen_data)

        self._send_ddp_data(byte_data)

    def _create_ddp_packet(self, data, data_offset, is_last_packet=False):
        header = bytearray(10)
        header[0] = 0b01000000 | (0b00000001 if is_last_packet else 0)
        header[1] = self.sequence_id
        header[2] = 0x01  # Data type set to 01
        header[3] = self.DDP_DESTINATION_ID
        header[4:8] = data_offset.to_bytes(4, byteorder="big")
        header[8:10] = len(data).to_bytes(2, byteorder="big")
        return header + data

    def _send_ddp_data(self, rgb_data):
        data_offset = 0

        for i in range(0, len(rgb_data), self.DDP_MAX_DATALEN):
            packet_data = rgb_data[i : i + self.DDP_MAX_DATALEN]
            is_last_packet = i + self.DDP_MAX_DATALEN >= len(rgb_data)
            packet = self._create_ddp_packet(packet_data, data_offset, is_last_packet)
            self.socket.sendto(packet, (self.host, 4048))
            data_offset += len(packet_data)

        # Increment sequence ID for the next RGB dataset
        self.sequence_id = (self.sequence_id + 1) % 16

    def close_socket(self):
        self.socket.close()



import wx

from .model import Box


def get_virtual_desktop_size() -> (int, int):
    total_width = 0
    total_height = 0
    for i in range(wx.Display.GetCount()):
        display = wx.Display(i)
        rect = display.GetGeometry()
        total_width += rect.GetWidth()
        total_height = max(total_height, rect.GetHeight())
    return total_width, total_height


class TransparentWindow(wx.Frame):
    border_size = 5

    def __init__(self, parent, title, pos, size, capture_box: Box):
        self.capture_box = capture_box
        pos = (pos[0] - self.border_size, pos[1] - self.border_size)
        size = (size[0] + 2 * self.border_size, size[1] + 2 * self.border_size)
        super().__init__(
            parent,
            title=title,
            style=wx.FRAME_SHAPED | wx.STAY_ON_TOP,
            pos=pos,
            size=size,
        )
        self.SetBackgroundStyle(wx.BG_STYLE_PAINT)  # Needed for shaped windows
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_SIZE, self.OnSize)  # Resize event
        self.CreateShapeBitmap(size)
        self.Bind(wx.EVT_LEFT_DOWN, self.OnMouseDown)
        self.Bind(wx.EVT_LEFT_UP, self.OnMouseUp)
        self.Bind(wx.EVT_RIGHT_DOWN, self.OnRightMouseDown)
        self.Bind(wx.EVT_RIGHT_UP, self.OnMouseUp)
        self.Bind(wx.EVT_MOTION, self.OnMouseMove)
        self.Bind(wx.EVT_LEAVE_WINDOW, self.OnMouseLeave)
        self.Bind(wx.EVT_ENTER_WINDOW, self.OnMouseEnter)

        self.dragging = False
        self.resizing = False
        self.dragStartPos = None
        self.max_x, self.max_y = get_virtual_desktop_size()

    def OnMouseDown(self, event):
        self.CaptureMouse()
        self.dragging = True
        self.dragStartPos = event.GetPosition()

    def OnRightMouseDown(self, event):
        self.CaptureMouse()
        self.resizing = True
        self.dragStartPos = event.GetPosition()

    def OnMouseUp(self, event):
        if self.HasCapture():
            self.ReleaseMouse()
        self.dragging = False
        self.resizing = False

    def OnMouseMove(self, event):
        if self.dragging:
            x, y = self.ClientToScreen(event.GetPosition())
            newpos = (
                max(0, min(self.max_x - self.GetSize()[0], x - self.dragStartPos.x)),
                max(0, min(self.max_y - self.GetSize()[1], y - self.dragStartPos.y)),
            )
            self.capture_box.left = newpos[0] + self.border_size
            self.capture_box.top = newpos[1] + self.border_size
            self.Move(newpos)
        elif self.resizing:
            x, y = self.ClientToScreen(event.GetPosition())
            newsize = (
                int(
                    self.GetSize()[0]
                    / self.GetSize()[1]
                    * max(1, y - self.GetPosition().y)
                ),
                max(1, y - self.GetPosition().y),
            )
            self.capture_box.width = newsize[0] - 2 * self.border_size
            self.capture_box.height = newsize[1] - 2 * self.border_size
            self.SetSize(newsize)

    def OnMouseLeave(self, event):
        self.SetCursor(wx.Cursor(wx.CURSOR_ARROW))

    def OnMouseEnter(self, event):
        self.SetCursor(wx.Cursor(wx.CURSOR_SIZENWSE))

    def CreateShapeBitmap(self, size):
        # Create a bitmap where the border is opaque and the center is transparent
        bitmap = wx.Bitmap(size[0], size[1], 32)
        dc = wx.MemoryDC(bitmap)
        dc.SetBackground(
            wx.Brush(wx.Colour(255, 0, 0, 1), wx.BRUSHSTYLE_TRANSPARENT)
        )  # Fully transparent brush
        dc.Clear()
        dc.SetPen(
            wx.Pen(wx.Colour(255, 0, 0, 1), self.border_size, wx.PENSTYLE_SOLID)
        )  # Red pen for the border
        dc.SetBrush(
            wx.Brush(wx.Colour(255, 0, 0, 1), wx.BRUSHSTYLE_TRANSPARENT)
        )  # Fully transparent brush
        dc.DrawRectangle(0, 0, size[0], size[1])
        del dc  # Need to delete the DC before setting shape
        self.SetShape(wx.Region(bitmap, wx.Colour(0, 0, 0, 0)))

    def OnPaint(self, event):
        # Custom drawing for the border
        dc = wx.PaintDC(self)
        width, height = self.GetClientSize()
        dc.SetBrush(wx.TRANSPARENT_BRUSH)  # Transparent brush for the interior
        pen = wx.Pen(wx.Colour(255, 0, 0, 1), self.border_size)
        dc.SetPen(pen)  # Red pen for the border
        dc.DrawRectangle(0, 0, width, height)
        pen.Destroy()

    def OnSize(self, event):
        # Recreate the shape bitmap when the window is resized
        self.CreateShapeBitmap(event.GetSize())
        event.Skip()







import dxcam
import cv2

from . import image_processor

camera = dxcam.create()

def capture_dxcam(capture_box, resolution):
    global camera
    rgb_array = camera.grab(region=(capture_box.left, capture_box.top, capture_box.left +capture_box.width, capture_box.top + capture_box.height))
    downscaled_array = cv2.resize(rgb_array, resolution, interpolation=cv2.INTER_AREA)
    rgb_array = image_processor.apply_filters_cv2(downscaled_array)

    return rgb_array


from dataclasses import dataclass

@dataclass
class Box:
    left: int
    top: int
    width: int
    height: int

    def getSize(self):
        return self.width, self.height

    def getCenter(self):
        return self.left + self.width // 2, self.top + self.height // 2

    def getTopLeft(self):
        return self.left, self.top

    def getPosition(self):
        return self.getTopLeft()
